@@ -68,7 +68,7 @@ class AsyncClient(client.Client):
             if not transports:
                 raise ValueError('No valid transports provided')
         self.transports = transports or valid_transports
-        self.queue, self.queue_empty = self._create_queue()
+        self.queue = self.create_queue()
         return await getattr(self, '_connect_' + self.transports[0])(
             url, headers, engineio_path)
 
@@ -111,8 +111,7 @@ class AsyncClient(client.Client):
             await self._send_packet(packet.Packet(packet.CLOSE))
             await self.queue.put(None)
             self.state = 'disconnecting'
-            if not abort:
-                await self.queue.join()
+            await self._trigger_event('disconnect', run_async=False)
             if self.current_transport == 'websocket':
                 await self.ws.close()
             if not abort:
@@ -149,6 +148,16 @@ class AsyncClient(client.Client):
         """
         return await asyncio.sleep(seconds)
 
+    def create_queue(self):
+        """Create a queue object."""
+        q = asyncio.Queue()
+        q.Empty = asyncio.QueueEmpty
+        return q
+
+    def create_event(self):
+        """Create an event object."""
+        return asyncio.Event()
+
     async def _connect_polling(self, url, headers, engineio_path):
         """Establish a long-polling connection to the Engine.IO server."""
         if aiohttp is None:  # pragma: no cover
@@ -165,20 +174,17 @@ class AsyncClient(client.Client):
                 'Connection refused by the server')
         if r.status != 200:
             raise exceptions.ConnectionError(
-                'Unexpected status code %s in server response', r.status)
+                'Unexpected status code {} in server response'.format(
+                    r.status))
         try:
             p = payload.Payload(encoded_payload=await r.read())
         except ValueError:
             six.raise_from(exceptions.ConnectionError(
                 'Unexpected response from server'), None)
-        open_packet = [pkt for pkt in p.packets
-                       if pkt.packet_type == packet.OPEN]
-        if open_packet == []:
+        open_packet = p.packets[0]
+        if open_packet.packet_type != packet.OPEN:
             raise exceptions.ConnectionError(
                 'OPEN packet not returned by server')
-        if len(p.packets) > 1:  # pragma: no cover
-            self.logger.info('extra packets found in server response')
-        open_packet = open_packet[0]
         self.logger.info(
             'Polling connection accepted with ' + str(open_packet.data))
         self.sid = open_packet.data['sid']
@@ -190,7 +196,10 @@ class AsyncClient(client.Client):
 
         self.state = 'connected'
         client.connected_clients.append(self)
-        await self._trigger_event('connect')
+        await self._trigger_event('connect', run_async=False)
+
+        for pkt in p.packets[1:]:
+            await self._receive_packet(pkt)
 
         if 'websocket' in self.upgrades and 'websocket' in self.transports:
             # attempt to upgrade to websocket
@@ -198,7 +207,7 @@ class AsyncClient(client.Client):
                 # upgrade to websocket succeeded, we're done here
                 return
 
-        self.start_background_task(self._ping_loop)
+        self.ping_loop_task = self.start_background_task(self._ping_loop)
         self.write_loop_task = self.start_background_task(self._write_loop)
         self.read_loop_task = self.start_background_task(
             self._read_loop_polling)
@@ -233,17 +242,46 @@ class AsyncClient(client.Client):
             else:
                 raise exceptions.ConnectionError('Connection error')
         if upgrade:
-            await ws.send(packet.Packet(packet.PING, data='probe').encode())
-            pkt = packet.Packet(encoded_packet=await ws.recv())
+            p = packet.Packet(packet.PING, data='probe').encode(
+                always_bytes=False)
+            try:
+                await ws.send(p)
+            except Exception as e:  # pragma: no cover
+                self.logger.warning(
+                    'WebSocket upgrade failed: unexpected send exception: %s',
+                    str(e))
+                return False
+            try:
+                p = await ws.recv()
+            except Exception as e:  # pragma: no cover
+                self.logger.warning(
+                    'WebSocket upgrade failed: unexpected recv exception: %s',
+                    str(e))
+                return False
+            pkt = packet.Packet(encoded_packet=p)
             if pkt.packet_type != packet.PONG or pkt.data != 'probe':
                 self.logger.warning(
                     'WebSocket upgrade failed: no PONG packet')
                 return False
-            await ws.send(packet.Packet(packet.UPGRADE).encode())
+            p = packet.Packet(packet.UPGRADE).encode(always_bytes=False)
+            try:
+                await ws.send(p)
+            except Exception as e:  # pragma: no cover
+                self.logger.warning(
+                    'WebSocket upgrade failed: unexpected send exception: %s',
+                    str(e))
+                return False
             self.current_transport = 'websocket'
+            if self.http:  # pragma: no cover
+                await self.http.close()
             self.logger.info('WebSocket upgrade was successful')
         else:
-            open_packet = packet.Packet(encoded_packet=await ws.recv())
+            try:
+                p = await ws.recv()
+            except Exception as e:  # pragma: no cover
+                raise exceptions.ConnectionError(
+                    'Unexpected recv exception: ' + str(e))
+            open_packet = packet.Packet(encoded_packet=p)
             if open_packet.packet_type != packet.OPEN:
                 raise exceptions.ConnectionError('no OPEN packet')
             self.logger.info(
@@ -256,10 +294,10 @@ class AsyncClient(client.Client):
 
             self.state = 'connected'
             client.connected_clients.append(self)
-            await self._trigger_event('connect')
+            await self._trigger_event('connect', run_async=False)
 
         self.ws = ws
-        self.start_background_task(self._ping_loop)
+        self.ping_loop_task = self.start_background_task(self._ping_loop)
         self.write_loop_task = self.start_background_task(self._write_loop)
         self.read_loop_task = self.start_background_task(
             self._read_loop_websocket)
@@ -273,7 +311,7 @@ class AsyncClient(client.Client):
             'Received packet %s data %s', packet_name,
             pkt.data if not isinstance(pkt.data, bytes) else '<binary>')
         if pkt.packet_type == packet.MESSAGE:
-            await self._trigger_event('message', pkt.data)
+            await self._trigger_event('message', pkt.data, run_async=True)
         elif pkt.packet_type == packet.PONG:
             self.pong_received = True
         elif pkt.packet_type == packet.NOOP:
@@ -294,7 +332,7 @@ class AsyncClient(client.Client):
 
     async def _send_request(
             self, method, url, headers=None, body=None):  # pragma: no cover
-        if self.http is None:
+        if self.http is None or self.http.closed:
             self.http = aiohttp.ClientSession()
         method = getattr(self.http, method.lower())
         try:
@@ -302,10 +340,6 @@ class AsyncClient(client.Client):
         except aiohttp.ClientError:
             return
 
-    def _create_queue(self):
-        """Create the client's send queue."""
-        return asyncio.Queue(), asyncio.QueueEmpty
-
     async def _trigger_event(self, event, *args, **kwargs):
         """Invoke an event handler."""
         run_async = kwargs.pop('run_async', False)
@@ -348,6 +382,7 @@ class AsyncClient(client.Client):
         interval.
         """
         self.pong_received = True
+        self.ping_loop_event.clear()
         while self.state == 'connected':
             if not self.pong_received:
                 self.logger.warning(
@@ -355,11 +390,15 @@ class AsyncClient(client.Client):
                 if self.ws:
                     await self.ws.close()
                 await self.queue.put(None)
-                self._reset()
                 break
             self.pong_received = False
             await self._send_packet(packet.Packet(packet.PING))
-            await self.sleep(self.ping_interval)
+            try:
+                await asyncio.wait_for(self.ping_loop_event.wait(),
+                                       self.ping_interval)
+            except (asyncio.TimeoutError,
+                    asyncio.CancelledError):  # pragma: no cover
+                pass
         self.logger.info('Exiting ping task')
 
     async def _read_loop_polling(self):
@@ -373,13 +412,11 @@ class AsyncClient(client.Client):
                 self.logger.warning(
                     'Connection refused by the server, aborting')
                 await self.queue.put(None)
-                self._reset()
                 break
             if r.status != 200:
                 self.logger.warning('Unexpected status code %s in server '
                                     'response, aborting', r.status)
                 await self.queue.put(None)
-                self._reset()
                 break
             try:
                 p = payload.Payload(encoded_payload=await r.read())
@@ -387,13 +424,22 @@ class AsyncClient(client.Client):
                 self.logger.warning(
                     'Unexpected packet from server, aborting')
                 await self.queue.put(None)
-                self._reset()
                 break
             for pkt in p.packets:
                 await self._receive_packet(pkt)
 
         self.logger.info('Waiting for write loop task to end')
         await self.write_loop_task
+        self.logger.info('Waiting for ping loop task to end')
+        self.ping_loop_event.set()
+        await self.ping_loop_task
+        if self.state == 'connected':
+            await self._trigger_event('disconnect', run_async=False)
+            try:
+                client.connected_clients.remove(self)
+            except ValueError:  # pragma: no cover
+                pass
+            self._reset()
         self.logger.info('Exiting read loop task')
 
     async def _read_loop_websocket(self):
@@ -404,15 +450,13 @@ class AsyncClient(client.Client):
                 p = await self.ws.recv()
             except websockets.exceptions.ConnectionClosed:
                 self.logger.warning(
-                    'WebSocket connection was closed, aborting')
+                    'Read loop: WebSocket connection was closed, aborting')
                 await self.queue.put(None)
-                self._reset()
                 break
             except Exception as e:
                 self.logger.info(
                     'Unexpected error "%s", aborting', str(e))
                 await self.queue.put(None)
-                self._reset()
                 break
             if isinstance(p, six.text_type):  # pragma: no cover
                 p = p.encode('utf-8')
@@ -421,6 +465,16 @@ class AsyncClient(client.Client):
 
         self.logger.info('Waiting for write loop task to end')
         await self.write_loop_task
+        self.logger.info('Waiting for ping loop task to end')
+        self.ping_loop_event.set()
+        await self.ping_loop_task
+        if self.state == 'connected':
+            await self._trigger_event('disconnect', run_async=False)
+            try:
+                client.connected_clients.remove(self)
+            except ValueError:  # pragma: no cover
+                pass
+            self._reset()
         self.logger.info('Exiting read loop task')
 
     async def _write_loop(self):
@@ -428,11 +482,15 @@ class AsyncClient(client.Client):
         pushed to the send queue.
         """
         while self.state == 'connected':
+            # to simplify the timeout handling, use the maximum of the
+            # ping interval and ping timeout as timeout, with an extra 5
+            # seconds grace period
+            timeout = max(self.ping_interval, self.ping_timeout) + 5
             packets = None
             try:
-                packets = [await asyncio.wait_for(self.queue.get(),
-                                                  self.ping_timeout)]
-            except self.queue_empty:
+                packets = [await asyncio.wait_for(self.queue.get(), timeout)]
+            except (self.queue.Empty, asyncio.TimeoutError,
+                    asyncio.CancelledError):
                 self.logger.error('packet queue is empty, aborting')
                 self._reset()
                 break
@@ -443,7 +501,7 @@ class AsyncClient(client.Client):
                 while True:
                     try:
                         packets.append(self.queue.get_nowait())
-                    except self.queue_empty:
+                    except self.queue.Empty:
                         break
                     if packets[-1] is None:
                         packets = packets[:-1]
@@ -462,7 +520,6 @@ class AsyncClient(client.Client):
                 if r is None:
                     self.logger.warning(
                         'Connection refused by the server, aborting')
-                    self._reset()
                     break
                 if r.status != 200:
                     self.logger.warning('Unexpected status code %s in server '
@@ -473,11 +530,11 @@ class AsyncClient(client.Client):
                 # websocket
                 try:
                     for pkt in packets:
-                        await self.ws.send(pkt.encode())
+                        await self.ws.send(pkt.encode(always_bytes=False))
                         self.queue.task_done()
                 except websockets.exceptions.ConnectionClosed:
                     self.logger.warning(
-                        'WebSocket connection was closed, aborting')
-                    self._reset()
+                        'Write loop: WebSocket connection was closed, '
+                        'aborting')
                     break
         self.logger.info('Exiting write loop task')
